
/* simple-file.cc */

// #include "clang/Basic/SourceManager.h"
// #include "clang/Driver/Compilation.h"
#include "clang/CodeGen/CodeGenAction.h"
#include "clang/Driver/Driver.h"
#include "clang/Frontend/CompilerInvocation.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/DiagnosticOptions.h"
// #include "clang/Frontend/FrontendActions.h"
// #include "clang/Frontend/FrontendDiagnostic.h"
#include "clang/Serialization/ASTReader.h"
#include "clang/Frontend/TextDiagnosticPrinter.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
// #include "llvm/ADT/OwningPtr.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
// #include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/System/Host.h"
#include "llvm/Target/TargetSelect.h"

// #include <stdio.h>
#include <iostream>
#include <stdexcept> // class std::runtime_error

using namespace std;

/* ---------------------------------------------------------------------- */

void info (string msg)
{
   cout << msg << endl;
}

void error (string msg)
{
   cerr << msg << endl;
   throw new std::runtime_error (msg);
}

/* ---------------------------------------------------------------------- */

void execute ()
{
   static const char* argv [] = { "program", "-x", "c++",
                                  // "-fsyntax-only",
                                  // "-include-pch", "_output/example4.pch",
                                  // "-include-pth", "output/tgbutton.pth",
                                  // "-I/opt/root-5.26.00/include",
                                  // "-Wl,-rpath", "-Wl,/opt/root-5.26.00/lib",
                                  // "-lGui",
                                  #if 0
                                  "-I/usr/lib/gcc/i586-manbo-linux-gnu/4.3.2/include",
                                  #endif
                                  "input/sample.cc",
                                  // "-"
                                  };

   static const int argc = sizeof (argv) / sizeof (argv[0]);

   clang::TextDiagnosticPrinter * diagClient =
      new clang::TextDiagnosticPrinter (llvm::errs(), clang::DiagnosticOptions ());

   clang::Diagnostic diags (diagClient);

   clang::driver::Driver driver ("simple",
                                 llvm::sys::getHostTriple(),
                                 "a.out",
                                 /* IsProduction = */ false,
                                 /* CXXIsProduction = */ false,
                                 diags);

   driver.setTitle ("clang interpreter");

   clang::CompilerInvocation * invocation = new clang::CompilerInvocation;
   clang::CompilerInvocation::CreateFromArgs
       (*invocation,
        argv + 1, // skip program name
        argv + argc,
        diags);

   // Create a compiler instance to handle the actual work.
   clang::CompilerInstance CI;
   CI.setLLVMContext (new llvm::LLVMContext);
   CI.setInvocation (invocation);

   // Create the compilers actual diagnostics engine.
   CI.createDiagnostics (argc, const_cast <char**> (argv));
   if (! CI.hasDiagnostics ())
      error ("No diagnostics");

   // Create and execute the frontend to generate an LLVM bitcode module.
   clang::CodeGenAction * action = new clang::EmitLLVMOnlyAction();
   if (! CI.ExecuteAction (*action))
      error ("Cannot execute action");

   llvm::Module * module = action->takeModule ();
   if (module == NULL)
      error ("No module");

   /* execute */

   llvm::InitializeNativeTarget ();

   std::string Error;
   llvm::ExecutionEngine * engine =
      llvm::ExecutionEngine::createJIT (module, &Error);
   if (engine == NULL)
     error ("Unable to make execution engine: " + Error);

   llvm::Function * func = module->getFunction ("main");
   if (func == NULL)
     error ("Function not found in module");

   std::vector <std::string> params;
   params.push_back (module->getModuleIdentifier ());

   engine->runFunctionAsMain (func, params, /* envp = */ NULL);

   /* clean up */

   // NO delete module;
   delete engine;
   delete action;
   // NO delete invocation;
   llvm::llvm_shutdown ();
}

/* ---------------------------------------------------------------------- */

// see: clang/examples/clang-interpreter/main.cpp
// see: clang/tools/driver/cc1_main.cpp

int main (int argc, const char * *argv)
{
   execute ();
   return 0;
}
