
/* simple-file.cc */

// #include "clang/Basic/SourceManager.h"
// #include "clang/Driver/Compilation.h"
#include "clang/CodeGen/CodeGenAction.h"
#include "clang/Driver/Driver.h"
#include "clang/Frontend/CompilerInvocation.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/DiagnosticOptions.h"
// #include "clang/Frontend/FrontendActions.h"
// #include "clang/Frontend/FrontendDiagnostic.h"
#include "clang/Serialization/ASTReader.h"
#include "clang/Frontend/TextDiagnosticPrinter.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
// #include "llvm/ADT/OwningPtr.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/System/Host.h"
#include "llvm/Target/TargetSelect.h"

// #include <stdio.h>
#include <iostream>
#include <stdexcept> // class std::runtime_error

using namespace std;

/* ---------------------------------------------------------------------- */

void info (string msg)
{
   cout << msg << endl;
}

void error (string msg)
{
   cerr << msg << endl;
   throw new std::runtime_error (msg);
}

/* ---------------------------------------------------------------------- */

void execute (const char * source, const char * source2)
{
   static const char* argv [] = { "program", "-x", "c++",
                                  #if 0
                                  "-I/usr/lib/gcc/i586-manbo-linux-gnu/4.3.2/include",
                                  #endif
                                  "string-input",
                                  };

   static const int argc = sizeof (argv) / sizeof (argv[0]);

   clang::TextDiagnosticPrinter * diagClient =
      new clang::TextDiagnosticPrinter (llvm::errs(), clang::DiagnosticOptions ());

   clang::Diagnostic diags (diagClient);

   clang::driver::Driver driver ("simple",
                                 llvm::sys::getHostTriple(),
                                 "a.out",
                                 /* IsProduction = */ false,
                                 /* CXXIsProduction = */ false,
                                 diags);

   driver.setTitle ("clang interpreter");

   clang::CompilerInvocation * invocation = new clang::CompilerInvocation;
   clang::CompilerInvocation::CreateFromArgs
       (*invocation,
        argv + 1, // skip program name
        argv + argc,
        diags);

   // Create a compiler instance to handle the actual work.
   clang::CompilerInstance CI;
   CI.setLLVMContext (new llvm::LLVMContext);
   CI.setInvocation (invocation);

   // Create the compilers actual diagnostics engine.
   CI.createDiagnostics (argc, const_cast <char**> (argv));
   if (! CI.hasDiagnostics ())
      error ("No diagnostics");

   // Create memmory buffer with source text
   llvm::MemoryBuffer * buffer = llvm::MemoryBuffer::getMemBufferCopy (source, "SIMPLE_BUFFER");
   if (buffer == NULL)
      error ("Failed to create memory buffer");

   // Remap auxiliary name "string-input" to memory buffer
   clang::PreprocessorOptions & opts = CI.getInvocation().getPreprocessorOpts();
   opts.addRemappedFile ("string-input", buffer);

   // Create and execute the frontend to generate an LLVM bitcode module.
   clang::CodeGenAction * action = new clang::EmitLLVMOnlyAction();
   if (! CI.ExecuteAction (*action))
      error ("Cannot execute action");

   llvm::Module * module = action->takeModule ();
   if (module == NULL)
      error ("No module");

   /* execute */

   llvm::InitializeNativeTarget ();

   std::string Error;
   llvm::ExecutionEngine * engine =
      llvm::ExecutionEngine::createJIT (module, &Error);
   if (engine == NULL)
     error ("Unable to make execution engine: " + Error);

   llvm::Function * func = module->getFunction ("main");
   if (func == NULL)
     error ("Function not found in module");

   std::vector <std::string> params;
   params.push_back (module->getModuleIdentifier ());

   engine->runFunctionAsMain (func, params, /* envp = */ NULL);

   /* second compilation */

   llvm::MemoryBuffer * buffer2 = llvm::MemoryBuffer::getMemBufferCopy (source2, "SECOND_BUFFER");
   if (buffer2 == NULL)
      error ("Failed to create memory buffer");

   // clang::PreprocessorOptions & opts = CI.getInvocation().getPreprocessorOpts();
   // opts.clearRemappedFiles ();
   opts.eraseRemappedFile(opts.remapped_file_buffer_end() - 1);
   opts.addRemappedFile ("string-input", buffer2);

   clang::CodeGenAction * action2 = new clang::EmitLLVMOnlyAction();
   if (! CI.ExecuteAction (*action2))
      error ("Cannot execute action");

   llvm::Module * module2 = action2->takeModule ();
   if (module2 == NULL)
      error ("No module");

   /* second execute */

   // llvm::ExecutionEngine * engine2 =
   //    llvm::ExecutionEngine::createJIT (module2, &Error);
   // if (engine2 == NULL)
   //   error ("Unable to make execution engine: " + Error);

   llvm::Function * func2 = module2->getFunction ("func2");
   if (func2 == NULL)
     error ("Function not found in module");

   engine->addModule (module2);

   engine->runFunctionAsMain (func2, params, /* envp = */ NULL);

   /* clean up */

   // NO delete module;
   delete engine;
   delete action;
   // NO delete invocation;
   llvm::llvm_shutdown ();
}

/* ---------------------------------------------------------------------- */

// see: clang/examples/clang-interpreter/main.cpp
// see: clang/tools/driver/cc1_main.cpp

int main (int argc, const char * *argv)
{
   const char * txt1 =
      "#include <stdio.h> \n"

      "extern \"C\" int subroutine (int a, int b) \n"
      "{ \n"
      "printf (\"Hello from subroutine, a = %d, b = %d \\n\", a, b);"
      "return a+b; \n"
      "} \n"

      "int main () \n"
      "{ \n"
      "int i = 1; \n"
      "i ++; \n"
      "i = subroutine (1, i); \n"
      "printf (\"Hello from main, i = %d \\n\", i);"
      "return 0; \n"
      "} \n" ;

   const char * txt2 =
      "#include <stdio.h> \n"
      "extern \"C\" int subroutine (int a, int b); \n"
      "extern \"C\" int func2 () \n"
      "{ \n"
      // "int i = subroutine (3, 4); \n"
      "int i = 1; \n"
      "printf (\"Hello from func2, i = %d \\n\", i);"
      "return 0; \n"
      "} \n";

   execute (txt1, txt2);
   return 0;
}
