
/* simple.cc */

#include "clang/AST/ASTConsumer.h"
#include "clang/AST/ASTContext.h"
#include "clang/AST/DeclGroup.h"
#include "clang/Basic/FileManager.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Basic/TargetInfo.h"
#include "clang/CodeGen/CodeGenAction.h"
#include "clang/CodeGen/ModuleBuilder.h"
#include "clang/Driver/Compilation.h"
#include "clang/Driver/Driver.h"
#include "clang/Frontend/CompilerInvocation.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/DiagnosticOptions.h"
#include "clang/Frontend/FrontendAction.h"
#include "clang/Frontend/FrontendActions.h"
#include "clang/Frontend/FrontendDiagnostic.h"
#include "clang/Serialization/ASTReader.h"
#include "clang/Frontend/TextDiagnosticPrinter.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Parse/ParseAST.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/GenericValue.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/System/Host.h"
#include "llvm/Target/TargetSelect.h"

#include <stdio.h>
#include <iostream>
#include <stdexcept> // class std::runtime_error

using namespace std;

/* ---------------------------------------------------------------------- */

void info (string msg)
{
   cout << msg << endl;
}

void error (string msg)
{
   cerr << msg << endl;
   throw new std::runtime_error (msg);
}

/* ---------------------------------------------------------------------- */

void openPCH (clang::CompilerInstance & CI, std::string fileName)
{
   clang::PreprocessorOptions & PO = CI.getInvocation().getPreprocessorOpts();

   std::string originalFile =
      #ifdef LLVM_2_8
      clang::ASTReader::getOriginalSourceFile(fileName, CI.getDiagnostics());
      #else
      clang::ASTReader::getOriginalSourceFile(fileName, CI.getFileManager(), CI.getFileSystemOpts(), CI.getDiagnostics());
      #endif

   if (! originalFile.empty())
   {
      PO.ImplicitPCHInclude = fileName;
      PO.Includes.push_back(originalFile);
   }
   else info ("Cannot find original file for precompiled header: " + fileName);
}

/* ---------------------------------------------------------------------- */

// #define SRC

void execute (string src)
{
   static const char* argv [] = { "program", "-x", "c++",
                                  // "-fsyntax-only",
                                  // "-include-pch", "output/example4.pch",
                                  // "-include-pth", "output/tgbutton.pth",
                                  // "-I/opt/root-5.26.00/include",
                                  // "-Wl,-rpath", "-Wl,/opt/root-5.26.00/lib",
                                  // "-lGui",
                                  #if 0
                                  "-I/usr/lib/gcc/i586-manbo-linux-gnu/4.3.2/include",
                                  #endif
                                  #ifdef SRC
                                  "input/mini.cc",
                                  #endif
                                  // "-"
                                 };

   static const int argc = sizeof (argv) / sizeof (argv[0]);

   clang::TextDiagnosticPrinter * diagClient =
      new clang::TextDiagnosticPrinter (llvm::errs(), clang::DiagnosticOptions ());

   clang::Diagnostic diags (diagClient);

   #if 0
   clang::driver::Driver driver ("simple",
                                 ".",
                                 llvm::sys::getHostTriple(),
                                 "a.out",
                                 /* IsProduction = */ false,
                                 /* CXXIsProduction = */ false,
                                 diags);

   driver.setTitle ("clang interpreter");
   #endif

   #if 0
   clang::driver::Compilation * compilation = driver.BuildCompilation (argc, argv);
   if (compilation == NULL)
      error ("No compilation");
   #endif

   clang::CompilerInvocation * invocation = new clang::CompilerInvocation;
   clang::CompilerInvocation::CreateFromArgs
      (*invocation,
       argv + 1,
       argv + argc,
       diags);

   // Create a compiler instance to handle the actual work.
   clang::CompilerInstance CI;
   CI.setLLVMContext (new llvm::LLVMContext);
   CI.setInvocation (invocation);

   // Create the compilers actual diagnostics engine.
   CI.createDiagnostics (argc, const_cast <char**> (argv));
   if (! CI.hasDiagnostics ())
      error ("No diagnostics");

   #if 0
   openPCH (CI, "_output/example4.pch");
   #endif

   /* ................................................................... */

   #ifndef SRC

      CI.setTarget
      (
         clang::TargetInfo::CreateTargetInfo
            (CI.getDiagnostics (),
             CI.getTargetOpts ())
      );

      if (!CI.hasTarget ())
         error ("No target");

      CI.getTarget().setForcedLangOptions(CI.getLangOpts());

      CI.createFileManager ();
      #ifdef LLVM_2_8
      CI.createSourceManager ();
      #else
      CI.createSourceManager (CI.getFileManager(), CI.getFileSystemOpts());
      #endif

      CI.createPreprocessor();
      clang::Preprocessor & PP = CI.getPreprocessor();

      CI.getDiagnosticClient ().BeginSourceFile (CI.getLangOpts(), &PP);

      CI.setASTContext
      (
         new clang::ASTContext
                (CI.getLangOpts (),
                 PP.getSourceManager (),
                 CI.getTarget (),
                 PP.getIdentifierTable (),
                 PP.getSelectorTable (),
                 PP.getBuiltinInfo (),
                 // false,
                 0)
      );

      CI.setASTConsumer (new clang::ASTConsumer());

      #if 0
      PP.getBuiltinInfo().InitializeBuiltins
         (PP.getIdentifierTable (),
          PP.getLangOptions().NoBuiltin);
      #endif

      #ifndef FROM_FILE
         llvm::MemoryBuffer * SB = llvm::MemoryBuffer::getMemBufferCopy(src, "SIMPLE_BUFFER");
         if (SB == NULL)
            error ("Failed to create memory buffer");

         // CI.getSourceManager().clearIDTables();
         CI.getSourceManager().createMainFileIDForMemBuffer(SB);
         if (CI.getSourceManager().getMainFileID().isInvalid())
            error ("compileString: Failed to create main file id");
      #else
         std:string fileName = "input/mini.cc";
         const clang::FileEntry* File = CI.getFileManager().getFile(fileName);
         if (File)
            CI.getSourceManager().createMainFileID(File, clang::SourceLocation());
         if (CI.getSourceManager().getMainFileID().isInvalid())
            error ("Error reading file");
      #endif

      clang::ParseAST (PP, & CI.getASTConsumer (), CI.getASTContext ());

      unsigned err_count = CI.getDiagnostics().getNumErrors();
      if (err_count != 0)
         error ("Parse failed");

      clang::TranslationUnitDecl* tu = CI.getASTContext().getTranslationUnitDecl();
      if (tu == NULL)
         error ("Parse failed, no translation unit");

   #endif // not SRC

   /* ................................................................... */

   #ifndef SRC

      /* generate code */

      llvm::OwningPtr <clang::CodeGenerator> codeGen
      (
         CreateLLVMCodeGen (CI.getDiagnostics (),
                            "ENTRY_MODULE",
                            CI.getCodeGenOpts (),
                            CI.getLLVMContext ())
      );

      codeGen->Initialize (CI.getASTContext ());

      info ("Running code generation");
      clang::TranslationUnitDecl::decl_iterator iter = tu->decls_begin ();
      clang::TranslationUnitDecl::decl_iterator iter_end = tu->decls_end ();
      for ( ; iter != iter_end; ++ iter )
         codeGen->HandleTopLevelDecl (clang::DeclGroupRef (*iter));
      codeGen->HandleTranslationUnit(CI.getASTContext ());
      info ("Finished code generation");

      llvm::Module* module = codeGen->ReleaseModule();
      if (module == NULL)
         error ("Module creation failed");

   #endif

   /* ................................................................... */

   #ifdef SRC

      #if 0
      clang::FrontendAction * action1 = new clang::SyntaxOnlyAction();
      if (! CI.ExecuteAction (*action1))
         error ("Cannot execute action");
      delete action1;
      #endif

      // Create and execute the frontend to generate an LLVM bitcode module.
      clang::CodeGenAction * action = new clang::EmitLLVMOnlyAction;
      if (! CI.ExecuteAction (*action))
         error ("Cannot execute action");

      llvm::Module * module = action->takeModule ();
      if (module == NULL)
         error ("No module");

      delete action;

   #endif

   /* ................................................................... */

   /* execute */

   llvm::InitializeNativeTarget ();

   std::string Error;
   llvm::ExecutionEngine * engine =
      llvm::ExecutionEngine::createJIT (module, &Error);
   if (engine == NULL)
     error ("Unable to make execution engine: " + Error);

   llvm::Function * func = module->getFunction ("main");
   if (func == NULL)
     error ("Function not found in module");

   std::vector <std::string> params;
   params.push_back (module->getModuleIdentifier ());

   engine->runFunctionAsMain (func, params, /* envp = */ NULL);

   /* ................................................................... */


   /* clean up */

   // NO delete module;
   delete engine;
   // NO delete invocation;
   // delete compilation;

   llvm::llvm_shutdown ();

}

/* ---------------------------------------------------------------------- */

// see: clang/examples/clang-interpreter/main.cpp
// see: clang/tools/driver/cc1_main.cpp

int main (int argc, const char * *argv)
{
   execute ("#include <stdio.h> \n"
            "int main () \n"
            "{ \n"
            "int i = 1; \n"
            "i ++; \n"
            #if 0
            "i = f3 (1, 2); \n"
            #endif
            "printf (\"Hello from string, i = %d \\n\", i);"
            "return 0; \n"
            "} \n");
   return 0;
}
